--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   MultivariatePol.hs
--
-- Author:
--   Klaus Aehlig (2010)
--
-- Description:

module MultivariatePol where

import RAMLTypes
import Util
import qualified Data.Map as M


-- Type Indices: Other than in the paper we work with tuples instead of pairs.
-- The only index for a base type is TIUnit.  An index for a list of type A is 
-- a list of indices for type A.  An index for an n-tuple is an n-tuple of
-- indices corresponding to the subtypes.

data TIndices = TIUnit | TIList [TIndices] | TITuple [TIndices] deriving Eq

instance Show TIndices where
    show TIUnit = "*"
    show (TIList xs) = if (allZero xs) then show (length xs) else show xs
    show (TITuple xs) = "(" ++ showWithSep "," xs ++ ")"

allZero xs = all (\x -> x == TIUnit
          || x == TIList []
          || case x of ; TITuple xs' -> allZero xs' ; _ -> False ) xs

-- for later use in Data.Map, define some arbitrary linear order on the indices
-- note that the order is actually unimportant, it is only important that it be
-- linear

lexOrdList []       _    = True
lexOrdList _       []    = False
lexOrdList (a:as) (b:bs) = if (a < b) then True else if (b < a) then False else lexOrdList as bs


instance Ord TIndices where
     TIUnit      <= _          = True
     TIList xs   <= TIList ys  = lexOrdList xs ys
     TIList _    <= _          = True
     TITuple xs  <= TITuple ys = lexOrdList xs ys
     _           <= _          = False


--
-- For a given type and maximal degree we can find the (finite!) set of valid indices
--


-- distributeWeight k l computes all possibilities to distribute k into |l| positions.
-- i.e. distributeWeight 3 [(),()] = [[0,3],[1,2],[2,1],[3,0]]
distributeWeight 0 [] = [[]]
distributeWeight _ [] = []
distributeWeight k (_:xs) = [0..k] >>= \k' -> map (\l -> k':l) (distributeWeight (k-k') xs)

-- combine computes the combination of all elements from a list of lists where
-- the order in one list is respected.
-- combine ["123","abc","DEF"] =
-- ["1aD","1aE","1aF","1bD","1bE","1bF","1cD","1cE","1cF","2aD",
-- ,"2aE","2aF","2bD","2bE","2bF","2cD","2cE","2cF","3aD","3aE"
-- ,"3aF","3bD","3bE","3bF","3cD","3cE","3cF"]
combine []            = [[]]
combine ([]:xs)       = []
combine ((a:as):xs)   = (map ((:)a) (combine xs)) ++ combine (as:xs)

indicesWithDegree k (TTuple xs) = distributeWeight k xs >>= \w -> map TITuple (combine (map2 indicesWithDegree w xs))

indicesWithDegree k (TList a)   = [0..k] >>= \ len -> distributeWeight (k-len) [1..len] 
                                         >>= \ w -> map TIList (combine (map (flip indicesWithDegree a) w))

indicesWithDegree k (TBTree a)  = indicesWithDegree k (TList a)

indicesWithDegree 0 TNil      = [TIList []]

indicesWithDegree 0 TLeaf      = [TIList []]

indicesWithDegree 0 _           = [TIUnit]
indicesWithDegree _ _           = []

indicesWithDegreeAtmost k a = [0..k] >>= flip indicesWithDegree a

-- compute the degree of an index

degree TIUnit       = 0
degree (TIList is)  = length is + sum (map degree is)
degree (TITuple is) = sum (map degree is)


--
-- Mulitivariate Polynomials are implemented as (finite) Maps
-- from Indices to coefficients. We're parametric in the coefficients,
-- as this might be concrete numbers or variables for generating constraints
--

data MPolynomial a = MPolynomial (M.Map TIndices a)

appendWithSeparator s []     = []
appendWithSeparator s [a]    = a
appendWithSeparator s (a:as) = a ++ s ++ (appendWithSeparator s as) 
 
instance Show a => Show (MPolynomial a) where
    show (MPolynomial m) = "< " ++ appendWithSeparator "; " (map (\k -> show k ++ " |-> " ++ show (m M.! k)) (M.keys m)) ++ ">"



freshPolynomial vars k t = MPolynomial (M.fromList (zip (indicesWithDegreeAtmost k t) vars))

i0 t = the (indicesWithDegree 0 t)
constcoeff (MPolynomial p) t = p M.! (i0 t)


