--------------------------------
-------Resource Aware ML--------
--------------------------------
-- File:
--   Parser.hs
--
-- Author:
--   Jan Hoffmann (2009)
--
-- Description: 
--   A parser for RAML programs.  It is based on GHCs parser combinator
--   parsec (http://legacy.cs.uu.nl/daan/parsec.html) and especially on
--   Parsec.Expr which is a module that can be used to autmatically 
--   build parsers for expressions with binary operations.
--  
--   This contains a description of the RAML syntax.  It is given as
--   comments in the section "The Parser".
--  
--   Use the function "parseProgram" to parse a RAML program.




module Parser where

import RAMLTypes

import Text.ParserCombinators.Parsec hiding (State)
import qualified Text.ParserCombinators.Parsec.Token as P
import Text.ParserCombinators.Parsec.Language( emptyDef )
import Text.ParserCombinators.Parsec.Expr



-----------------------------------------------------------
-- External Interface
-----------------------------------------------------------

parseProgram :: (Monad m) => SourceName -> String -> m Program
{-
Input:   a filename (for error reporting) and a RAML program as a string
Output:  A program or a ParseError

A program is a pair of
- any list of function definitions and type declarations (it is not
  checked if the list is well-formed in any way)
- and a main expression

It fails with a parse error if the parser fails.
-}


parseProgram fName programString =  
    case parse program fName programString of
      Left err -> fail $ show err
      Right prog -> return prog
 


-- Just for the testing.
parseExpression exprStr =
    case parse expr exprStr exprStr of
      Left err -> fail $ show err
      Right e -> return e



-----------------------------------------------------------
-- Pretty Printing
-----------------------------------------------------------


prettyRAMLFromFile :: FilePath -> IO ()
{-
A pretty printer for RAML programs.  Can be used for testing.

Input:  a filename of a RaML program
Output: () in the IO monad

It prints the AST of the program if the parser was succesfull
or it prints the error message of the parser if not.
 
-}


prettyRAMLFromFile fname
  = do{ input <- readFile fname
      ; putStr input
      ; putStr "\n\n\n"
      ; case parse program fname input of
           Left err -> do{ putStr "parse error at "
                           ; print err
                           }
           Right x  -> print x
      }




-----------------------------------------------------------
-- The Parser
-----------------------------------------------------------

{-
It follows the parser combinator together with the description
of the RAML syntax.
-}



program :: Parser Program

program = 
   do{ whiteSpace                          -- A RAML Program is:
     ; defs <- many (lexeme declaration)   -- - a list of 0 or more declarations  
     ; reserved "main"                     -- - the word "main"
     ; reservedOp "="                      --   followed by "="
     ; e <- expr                           --   followed by an expression
     ; char ';' <|> (return ';')           --   an optional ";"
     ; whiteSpace
     ; eof                                 --   and nothing else
     ; return (defs,e)
     }



declaration :: Parser Dec

declaration =                       -- A declaration is:
   do { pos <- getPosition          -- (The position in the file is stored for error reporting.)
      ; funname <- identifier       -- an identifier
        <?> "function name"
      ;     typedec funname pos     -- followed by a type declaration
        <|> fundec funname pos      -- or a function definition
      }

   where
      typedec name pos =           -- A type declaration is
         do { reservedOp ":"       -- a colon
            ; args <- typ          -- a simple type (the argument type)
            ; reservedOp "->"      -- the functin arrow
            ; result <- typ        -- another simple type (the result type)
            ; return $ TypeDec name (FunType args result) pos
            }

      fundec name pos =                        -- A function definition is
         do { args <- parens (commaSep1 varId) -- parentheses enclosed list of variable identifiers
                      <|> do { x <- varId      -- or just a variable identifier
                             ; return [x] 
                             }
            ; reservedOp "="                   -- followed by "=",
            ; e <- expr                        -- an expression
            ; char ';'                         -- and an semicolon (which is needed to separate a variable
            ; return $ FunDec name args e pos  -- from the next declaration) (this sucks)
            }


typ :: Parser Type

typ =                                          -- A type is
       do { reserved "int"                     -- the strings "int"
          ; return TInt
          }
   <|> do { reserved "bool"                    -- or the string "bool"
          ; return TBool
          }
   <|> do { reserved "unit"                    -- or the string "unit"
          ; return TUnit
          }
   <|> do { types <- parens (commaSep1 typ)    -- or a non-empty parentheses enclosed list of types
          ; return $ mkTuple types             -- which is a tuple iff there is more then one element
          }
   <|> do { reserved "L"                       -- or the string "L"
          ; types <- parens (commaSep1 typ)    -- followed a non-empty parentheses enclosed list of types
          ; return (TList (mkTuple types)) }   -- that is conv. into a tuple iff there is more then one elem
   <|> do { reserved "T"                       -- or the string "T"
          ; types <- parens (commaSep1 typ)    -- followed a non-empty parentheses enclosed list of types
          ; return (TBTree (mkTuple types)) }  -- that is conv. into a tuple iff there is more then one elem
   <?> "type"
   where
     mkTuple [t]    = t
     mkTuple ts = TTuple ts



varId = identifier <?> "variable name"         -- a varId is just an identifier that gives a nice error
                                               -- message if it fails


expr1 :: Parser Exp
{-
expr1 parses an expression without (binary) operations.
It is then used to build the parser expr from a list
of operations.  See below.
-}
expr1 =                                                -- An expression without (bin) ops is:
   do { pos <- getPosition
      ; choice                                         -- (choice is the same as <|>'s: first match is taken)
         [ do { reserved "true"                        -- the string "true"
              ; return $ ETrue pos
              }
         , do { reserved "false"                       -- the string "false"
              ; return $ EFalse pos
              }
         , do { n <- lexeme decimal                    -- an decimal number without leading "-" or "+"
              ; return $ EInt (fromIntegral n) pos                    -- the latter cause confusion with the ops + and -
              }
         , do { x <- identifier                        -- an identifier (outfactor the identifier)
              ; do { arg <- expr                       -- followod by an expression
                   ; return $ EApp x arg pos           -- which means that it is a function application
                   }                                   -- or follows by something else
                <|> return (EVar x pos)                -- which means that it is just a variable
              }
         , letExp pos                                  -- a let expression starting with "let"
         , conditional pos                             -- a conditional starting with "if"
         , do { reserved "nil"                         -- the string "nil"
              ; return $ ENil pos
              }
         , do { reserved "cons"                        -- the string "cons"
              ; (e1,e2) <- pairOf expr                 -- followed by a pair of expressions (see below)
              ; return $ ECons e1 e2 pos
              }
         , match pos                                   -- a match expression starting with "match"
         , do { es <- parens (commaSep expr)           -- a parentheses enclosed (maybe empty) list of expr.
              ; case es of                             -- that is
                  []  -> return $ EUnit pos            -- just the constant unit
                  [e] -> return e                      -- an expression in parentheses
                  _   -> return $ ETuple es pos        -- or a tuple of expressions
              }
         , do { es <- squares (commaSep expr)          -- a square brackets enclosed list of expressions
              ; return $ foldr (\h t -> ECons h t pos) -- that are composed via cons nodes
                               (ENil pos)              -- starting with nil
                               es
              }
         , do { reserved "leaf"                        -- the string "leaf"
              ; return $ ELeaf pos
              }
         , treeNode pos                                -- a tree node starting with "node"
         ]
       }
   where
      treeNode pos =                                   -- A tree node for a binary tree or an (unrestricted) tree
         do { reserved "node"                          -- starts with the word "node"
            ; symbol "("                               -- followed by a list of expressions
            ; e1 <- expr
            ; comma
            ; e2 <- expr
            ; comma                                    -- If there is a triple of expressions then
            ; e3 <- expr                               -- it is a binary tree node
            ; symbol ")"
            ; return $ ENodeB e1 e2 e3 pos
            }

      letExp pos =                                     -- A let expression is either a "normal" let or a
         do { reserved "let"                           -- "tuple match".  It starts with "let"
            ;     do { x <- varId                      -- a "normal" let continues with a variable ident.
                     ; reservedOp "="                  -- followed by "=",
                     ; e1 <- expr                      -- an expression,
                     ; reserved "in"                   -- the string "in"
                     ; e2 <- expr                      -- and another expression
                     ; return $ ELet x e1 e2 pos
                     }
              <|> do { xs <- varTuple                  -- a "tuple match" starts with a parents enclosed
                     ; reservedOp "="                  -- varialbe list of length >1, followed by "=",
                     ; e1 <- expr                      -- an expression,
                     ; reserved "in"                   -- the string "in"
                     ; e2 <- expr                      -- and another expression
                     ; return $ EMatchT e1 xs e2 pos   -- which is a way of writing match for tuples
                     }
            }
      
      conditional pos =                                -- A conditional is:
         do { reserved "if"                            -- the string "if"
            ; e <- expr                                -- followed by an expression,
            ; reserved "then"                          -- the string "then",
            ; e1 <- expr                               -- another expression,
            ; reserved "else"                          -- the string "else"
            ; e2 <- expr                               -- and a third expression
            ; return $ ECond e e1 e2 pos
            }
      
      
      match pos =                                      -- A match (list or tuple) expression is:
         do { matchType <- do { reserved "matchD"      -- the string "match"
                              ; return Dest            -- it is either a destructive match ("matchD")
                              }
                       <|> do { reserved "match"       -- or a non-destructive match (just "match")
                              ; return NonDest
                              }
            ; e <- expr                                -- followed by and expression
            ; reserved "with"                          -- and the string "with |".
            ; reservedOp "|"                           -- It can be either a list match or a tuple match.

            ; try (do { vars <- varTuple               -- A tuple match starts with a parent encl. list of vars (> 1 var)
                      ; reservedOp "->"                -- followed by "->"
                      ; e' <- expr                     -- and an expression.
                      ; return $ EMatchT e vars e' pos -- (try is needed since it could also be a cons match for lists)
                      } )

          <|> do { reserved "nil"                      -- A list match starts with the string "nil"
                   <|> (squares $ return ())           -- or with "[]"
                 ; reservedOp "->"                     -- followed by "->",
                 ; e1 <- expr                          -- an expression,
                 ; reservedOp "|"                      -- "|",
                 ; (x1,x2) <- consMatch                -- something like (x::xs) or cons(x,xs)
                 ; reservedOp "->"                     -- Then again "->"
                 ; e2 <- expr                          -- and another expression.
                 ; return $ EMatchL matchType e e1 x1 x2 e2 pos
                 }                                     -- Alternatively, the cons-match comes first:

          <|> do { (x1,x2) <- consMatch                -- something like (x::xs) or cons(x,xs)
                 ; reservedOp "->"                     -- then "->",
                 ; e2 <- expr                          -- an expression,
                 ; reservedOp "|"                      -- "|",
                 ; reserved "nil"                      -- the string "nil",
                   <|> (squares $ return ())           -- or "[]"
                 ; reservedOp "->"                     -- followed by "->",
                 ; e1 <- expr                          -- and an expression.
                 ; return $ EMatchL matchType e e1 x1 x2 e2 pos
                 }

          <|> do { reserved "leaf"
                 ; reservedOp "->"
                 ; e1 <- expr
                 ; reservedOp "|"
                 ; reserved "node"
                 ; (x1,x2,x3) <- tripleOf varId
                 ; reserved "->"
                 ; e2 <- expr
                 ; return $ EMatchB matchType e e1 x1 x2 x3 e2 pos
                 }

          <|> do { reserved "node"
                 ; (x1,x2,x3) <- tripleOf varId
                 ; reserved "->"
                 ; e2 <- expr
                 ; reservedOp "|"
                 ; reserved "leaf"
                 ; reservedOp "->"
                 ; e1 <- expr
                 ; return $ EMatchB matchType e e1 x1 x2 x3 e2 pos
                 }

          <?> "nil or tuple"
            }

      
      
      
      pairOf parser = do { symbol "("
                         ; x1 <- parser
                         ; comma
                         ; x2 <- parser
                         ; symbol ")"
                         ; return (x1,x2) }

      tripleOf parser = do { symbol "("
                           ; x1 <- parser
                           ; comma
                           ; x2 <- parser
                           ; comma
                           ; x3 <- parser
                           ; symbol ")"
                           ; return (x1,x2,x3) }
      
      consMatch = do { symbol "("
                     ; x1 <- varId
                     ; reservedOp "::"
                     ; x2 <- varId
                     ; symbol ")"
                     ; return (x1,x2) }
              <|> do { reserved "cons"    -- "cons"
                     ; pairOf varId       -- and a pair of variables
                     }

      varTuple = 
         parens (
            do { x <- varId 
               ; comma
               ; xs <- commaSep1 varId
               ; return $ x:xs
               } )
         <?> "Tuple"
      

expr :: Parser Exp
{-
The parser for RAML expressions.  It is built from expr1
and the operations in the table below.

The table works as follows: It is a list of list such that the
operatoins in the top-most list have (the same) highest precedences
and so on.

Each sublist contains entries of the form
   Prefix parser
where parser is a parser that parses a (prefix) operator and returns
a function that computes something with the following expression (expr).
Or entries of the form
 Infix parser assoc
where assoc is a associativity and parser is a parser that parsers a 
(infix) operator and returns a function that computes something with
the expressions on the left and right.
-}

expr = buildExpressionParser table expr1

   where

      table = 
          [ [ binary "*" (\pos e1 e2 -> EBinOp Mult e1 e2 pos) AssocLeft]
          , [ binary "+" (\pos e1 e2 -> EBinOp Add e1 e2 pos) AssocLeft
            , binary "-" (\pos e1 e2 -> EBinOp Sub e1 e2 pos) AssocLeft
            ]
          , [ binary "mod" (\pos e1 e2 -> EBinOp Mod e1 e2 pos) AssocLeft
            , binary "div" (\pos e1 e2 -> EBinOp Div e1 e2 pos) AssocLeft
            ]
          , [ binary "==" (\pos e1 e2 -> EBinOp Eq e1 e2 pos) AssocLeft
            , binary "<=" (\pos e1 e2 -> EBinOp LessEq e1 e2 pos) AssocLeft
            , binary ">=" (\pos e1 e2 -> EBinOp GreaterEq e1 e2 pos) AssocLeft
            , binary "<" (\pos e1 e2 -> EBinOp Less e1 e2 pos) AssocLeft
            , binary ">" (\pos e1 e2 -> EBinOp Greater e1 e2 pos) AssocLeft
            ]
          , [ prefix "not" (\pos e -> ENot e pos) ]
          , [ binary "and" (\pos e1 e2 -> EBinOp And e1 e2 pos) AssocLeft
            , binary "or" (\pos e1 e2 -> EBinOp Or e1 e2 pos) AssocLeft
            ]
          , [ binary "::" (\pos e1 e2 -> ECons e1 e2 pos) AssocRight ]
          ]
            
      prefix str result = 
            Prefix (do { pos <- getPosition
                       ; reservedOp str
                       ; return (result pos) 
                       } )


      binary str result assoc = 
            Infix (do { pos <- getPosition
                      ; reservedOp str
                      ; return (result pos) 
                      } ) 
                  assoc



-----------------------------------------------------------
-- The lexer
-----------------------------------------------------------
{-

Following the parsec documentation, we build a lexer.  It is used to
distinguish between identifiers, reserved words and operators.

See the parsec documentation for the semantics of the lexer.
-}


lexer = P.makeTokenParser ramlDef

   where 
     ramlDef = emptyDef
           { P.commentStart   = "(*"
           , P.commentEnd     = "*)"
           , P.commentLine    = ""     -- no line comments in sML
           , P.nestedComments = True
           , P.identStart     = letter <|> char '_' 
           , P.identLetter    = alphaNum <|> oneOf "_'"
           , P.opStart        = oneOf ":!#$%&*+./<=>?@\\^|-~"
           , P.opLetter       = oneOf ":!#$%&*+./<=>?@\\^|-~"
           , P.reservedOpNames= ["::","|","=",":","->","<","<=",">",">=","==","+","-","*"]
           , P.reservedNames  = ["maxdegree", "main","true","false","let","in","if"
                                ,"then","else","match","matchD","with","nil","cons","unit"
                                ,"bool","int","L","T","and","or","not","mod","div"
                                ,"leaf","node"
                                ]
           , P.caseSensitive  = True
           }

lexeme          = P.lexeme lexer 
parens          = P.parens lexer    
commaSep1       = P.commaSep1 lexer    
commaSep        = P.commaSep lexer    
semiSep         = P.semiSep lexer    
whiteSpace      = P.whiteSpace lexer    
identifier      = (P.identifier lexer) <?> "identifier"
reserved        = P.reserved lexer    
reservedOp      = P.reservedOp lexer
decimal         = P.decimal lexer    
comma           = P.comma lexer
semi            = P.semi lexer
squares         = P.squares lexer
symbol          = P.symbol lexer    
charLiteral     = P.charLiteral lexer    
stringLiteral   = P.stringLiteral lexer